#if !defined (OGFACCE_H)
#define OGFACCE_H

#include "defs.h"
#include "math.h"

class ografo;

//la faccia e' chiusa o aperta. Ogni vertice su cui incide e' contato una sola volta.
//la faccia e' aperta se e solo se lo spigolo tra l'ultimo vertice e il primo ha tor==3
class faccia {
  std::vector<int> passaggio;
  std::vector<int> uscita;       //uscita[i] e' la direzione con cui esce da passaggio[i]
  std::vector<int> entrata;
  mutable int colore;
 public:
  faccia():passaggio(), uscita(), entrata(), colore(0) {}
	faccia(int i):passaggio(i,0), uscita(i,0), entrata(i,0), colore(0) {}

  int lati() const;          //ritorna il numero di vertici adiacenti
  bool embedded() const;    //ritorna 1 se chiuso e embedded
  void inverti();
  void ruota(int n);         //ruota di n: il vecchio passaggio[n] diventa il nuovo passaggio[0]

  int get_passaggio(int i) const {return passaggio[i];}
  int get_uscita(int i) const {return uscita[i];}
  int get_entrata(int i) const {return entrata[i];}
  int get_colore() const {return colore;}
  void set_uscita(int i, int j) {uscita[i]=j;}
  void set_entrata(int i, int j) {entrata[i]=j;}
  void set_passaggio(int i, int j) {passaggio[i]=j;}
  void set_colore(int c) const {colore = c;}

	std::vector<int> get_passaggi() {return passaggio;}
  void add_passaggio(int p, int u, int e);

	friend std::vector<faccia> facce(const ografo& o);
};

class teta {
  int elem[3];
 public:
	teta(const std::vector<faccia>& ff, int pt, int dir);
  teta(int a=0,int b=0,int c=0) {elem[0]=a; elem[1]=b; elem[2]=c;}

  int get_elem(int i) const {return elem[i];}
  void set_elem(int i, int j) {elem[i]=j;}

  int somma() const {return (elem[0]+elem[1]+elem[2]);}
  int prodotto() const {return (elem[0]*elem[1]*elem[2]);}
  bool contiene(int i) const {return (elem[0]==i || elem[1]==i || elem[2]==i);}
  bool nullo() const;
  bool ammissibile() const;
  bool ammissibile(int ) const;
  bool counterpass() const;
  bool pass_twice() const;
  int faccia_sola() const;
  bool triple_counterpass() const;
  void ordina() {::ordina(elem, 2);}
  void ruota(int);
  void scrivi(std::ostream&) const;
  friend teta operator-(teta);
  friend bool operator==(teta, teta);
  friend teta operator% (const teta& t, int modulo);
  friend teta abs(teta);
  friend bool short_loop(teta, teta);
  std::vector<int> facce() const;
};

class tetraedro{
  int elem[6];
 public:
  tetraedro(int A=0,int B=0,int C=0,int D=0,int E=0,int F=0) 
    {elem[0]=A; elem[1]=B; elem[2]=E; elem[3]=C; elem[4]=D; elem[5]=F; }
  tetraedro(const std::vector<faccia>&, int);
  int get_elem(int i) const {return elem[i];}
  void set_elem(int i, int j) {elem[i]=j;}
  int somma() const;
	int max_value() const;
  friend tetraedro operator% (const tetraedro&, int modulo);
};

void colora(const std::vector<faccia>&, const std::vector<int>);
void colora_progressivamente(const std::vector<faccia>&);

void ordina_da_spigolo(std::vector<faccia> ff, int vert, int col);

#endif
