#include <stdlib.h>
#include <math.h>            
#include <string>
#include <iostream>
#include <list>
#include <vector>
#include <algorithm>
#include <functional>

#include "defs.h"
#include "math.h"
#include "string.h"
#include "ograph.h"
#include "surfaces.h"
#include "groups.h"


typedef	std::list<relazione>::const_iterator relazione_const_iterator;
typedef	std::list<relazione>::iterator relazione_iterator;
typedef	std::vector<int>::const_iterator int_const_iterator;
typedef	std::vector<int>::iterator int_iterator;


//restituisce una presentazione del gruppo fondamentale dell'ografo o
gruppo pi(const ografo& o) {
  const int l=o.lung();
  tensor2 gen = tensor2(l,std::vector<int> (4,0)); 
	const int generatori = l+1;

	//costruisce albero generatore
  for (int i=0; i<l; i++) {
    for (int j=0; j<4; j++) {
      gen[i][j]=1; gen[o.get_v(i,j)][o.get_col(i,j)]=1;
      if (!o.connesso(gen)) {
				gen[i][j]=0; gen[o.get_v(i,j)][o.get_col(i,j)]=0;
      }
    }
  }   

	//enumera (partendo da 1) gli spigoli nel complementare dell'albero (con segno)
  int contatore=1;
  for (int i=0; i<l; i++) {
    for (int j=0; j<4; j++) {
      if (gen[i][j]==1) {
				gen[i][j]=++contatore;
				gen[o.get_v(i,j)][o.get_col(i,j)]=0-contatore;
      }
    }
  }	
  for (int i=0; i<l; i++) 
		for (int j=0; j<4; j++) 
			if (gen[i][j]) gen[i][j]=gen[i][j]-1+2*(gen[i][j]<0);

	//costruisce le relazioni dalle facce
	gruppo ris = gruppo(generatori);
  for (int j=0; j<o.get_ff().size(); j++) {
		const faccia& f = o.get_ff()[j];
		relazione nuova = relazione();
    for (int i=0; i<f.lati(); i++) {
      int e = gen[f.get_passaggio(i)][f.get_uscita(i)];
      if (e) nuova.push_back(e);
    }
		ris.relazioni.push_back(nuova);
  }
	return ris;
}

//servono a 'abelianizzato'
bool ridotto(tensor2&, int, int);
bool riduci(tensor2&, int&, int&);
int complessita(tensor2&, int, int);

//restituisce l'abelianizzato di un gruppo
gruppo_abeliano abelianizzato(const gruppo& g) {
  int generatori = g.num_generatori();
	int relazioni = g.num_relazioni();

	//costruisce una matrice w che determina il quoziente
	tensor2 w = tensor2(relazioni+1, std::vector<int>(generatori+1,0));
	relazione_const_iterator rl = g.relazioni.begin();
  for (int i=0; i<relazioni; i++, rl++)
		for (int j=0; j<rl->lung(); j++) 
      w[i][abs(rl->get_rel(j))] += (1-2*(rl->get_rel(j)<0));
	
	//da qui in poi c'e' probabilmente da sfoltire
	gruppo_abeliano ris = gruppo_abeliano();
  unsigned int coeff=1;
  int lungo=-1; bool segno=0;
  while (!ridotto(w,relazioni,generatori) && (lungo<0 || lungo>complessita(w,relazioni,generatori)) || segno) {
    segno=0;
    lungo = complessita(w,relazioni,generatori);
    int coeff_mom=0;
    for (int i=0; i<relazioni; i++)
      for (int j=1; j<=generatori; j++)
				if (w[i][j]) coeff_mom = coeff_mom = (coeff_mom==0) ? abs(w[i][j]) : ::MCD(coeff_mom, w[i][j]);
    if (coeff_mom==0) coeff_mom=1;
    if (coeff_mom>1)
      for (int i=0; i<relazioni; i++)
				for (int j=1; j<=generatori; j++)
					w[i][j]/=coeff_mom;
    coeff*=coeff_mom;
    for (int j=1; j<=generatori; j++) {
      for (int j2=1; j2<=generatori; j2++) {
				if (j!=j2) {
					int lung; int lung2;
					do {
						lung=complessita(w,relazioni,generatori);
						for (int k=0; k<relazioni; k++)
							w[k][j2]+=w[k][j];
						lung2 = complessita(w,relazioni,generatori);
					} while (lung2<lung);
					do {
						lung=complessita(w,relazioni,generatori);
						for (int k=0; k<relazioni; k++)
							w[k][j2]-=w[k][j];
						lung2 = complessita(w,relazioni,generatori);
					} while (lung2<lung);
					for (int k=0; k<relazioni; k++)
						w[k][j2]+=w[k][j];
				}
      }
    }
    int prima = generatori; segno=riduci(w,relazioni,generatori);
    if (prima>generatori && coeff>=2)
      for (int j=0; j<prima-generatori; j++)
				ris.p.push_back(coeff);
  }
  if (ridotto(w,relazioni,generatori)) {
    for (int j=1; j<=generatori; j++) {
      int M=0;
      for (int i=0; i<relazioni; i++)
				if (w[i][j]) {
					w[i][j]*=coeff; w[i][j]=abs(w[i][j]);
					M = M ? MCD(M,w[i][j]) : w[i][j];
				}
      if (M!=1)
				ris.p.push_back(M);
    }
  }
	return ris;
}

//serve a gruppo_abeliano abelianizzato
bool ridotto(tensor2& w, int nr, int generatori) {
  for (int i=0; i<nr; i++) {
    int quanti=0;
    for (int j=1; j<=generatori; j++)
      if (w[i][j]) quanti++;
    if (quanti>=2) return FALSE;
  }
  return TRUE;
}

//serve a gruppo_abeliano abelianizzato
bool riduci(tensor2& w, int& nr, int& generatori) {
  for (int i=0; i<nr; i++) {
    int quanti=0;
    for (int p=1; p<=generatori && !quanti; p++)
      quanti+=abs(w[i][p]);
    if (!quanti) {
      for (int j=i; j<nr-1; j++)
				for (int p=1; p<=generatori; p++)
					w[j][p]=w[j+1][p];
      nr--;
    }
  }
  int sign=0;
  for (int p=1; p<=generatori && !sign; p++) {
    unsigned int M=0;
    for (int i=0; i<nr; i++)
      if (w[i][p]) (M==0) ? M=abs(w[i][p]) : M=::MCD(M,w[i][p]);
    if (M==1) {
      for (int i=0; i<nr; i++)
				if (w[i][p]<0) for (int j=1; j<=generatori; j++) w[i][j]*=-1;
      int sign=0;
      do {
				sign=0;
				for (int i=0; i<nr-1; i++)
					if (w[i][p]) {
						int i2=i+1;
						for (; !w[i2][p] && i2<nr; i2++) {}
						if (i2<nr) {
							int j1=i; int j2=i2; if (w[j1][p]>w[j2][p]) {int t=j1; j1=j2; j2=t;}
							for (int j=1; j<=generatori; j++)
								w[j2][j]-=w[j1][j];
							sign=1;
						}
					}
      } while (sign);
      int i=0;
      for (; i<nr && !w[i][p]; i++) {}
      for (int j=i; j<nr-1; j++)
				for (int k=1; k<=generatori; k++)
					w[j][k]=w[j+1][k];
      for (int i=0; i<nr; i++)
				for (int j=p; j<generatori; j++)
					w[i][j]=w[i][j+1];
      nr--; generatori--;
    }
    if (M==1) sign=1;
  }
  if (sign) riduci(w, nr, generatori);
  return sign;
}

//serve a gruppo_abeliano abelianizzato
int complessita(tensor2& w, int nr, int generatori) {
  unsigned int megamcd=0;
  for (int j=1; j<=generatori; j++) {
    unsigned int mcd=0;
    for (int k=0; k<nr; k++)
      if (w[k][j])
				mcd = (mcd==0) ? abs(w[k][j]) : ::MCD(mcd,w[k][j]);
    if (mcd==1) return 0;
    if (mcd) megamcd = (megamcd==0) ? mcd : ::MCD(megamcd,(int) mcd);
  }
  if (megamcd!=1) return 0;
  int ris=0;
  for (int i=0; i<nr; i++)
    for (int j=1; j<=generatori; j++)
      ris+=abs(w[i][j]);
  return ris;
}

//elimina dalla presentazione (ciclicamente) le coppie del tipo a*a^(-1)
void relazione::cancella_coppie_opposte() {
  for (int i=0; i+1<lung(); )
    if (rel[i] == -rel[i+1]) {
			rel.erase(rel.begin() + i, rel.begin() + i + 2);
			cancella_coppie_opposte();
		}
		else i++;
	if (lung() && rel.back() == -rel.front()) {
		rel.erase(rel.begin());
		rel.erase(rel.end()-1);
		cancella_coppie_opposte();
	}
}

//trasforma r in r^(-1)
void relazione::inverti() {
	std::reverse(rel.begin(), rel.end());
	std::transform(rel.begin(), rel.end(), rel.begin(), std::negate<int>());
}

//ruota r di n. Ad esempio se n=2, 'abcade' diventa 'cadeab'
void relazione::ruota(int n) {
	std::rotate(rel.begin(), rel.begin()+n, rel.end());
}

//sostituisce ogni occorrenza di e con sost, e di -e con sost^(-1)
//alla fine 'cancella_coppie_opposte', cioe' elimina coppie del tipo a*a^(-1)
void relazione::sostituisci(int e, const std::vector<int>& sost) {
	std::vector<int> temp = sost; 
	std::reverse(temp.begin(), temp.end());
	std::transform(temp.begin(), temp.end(), temp.begin(), std::negate<int>());
	const std::vector<int> sost_inversed = temp;

	//fa schifo, ma non ho trovato niente di meglio:
	std::vector<int> abs_sost = abs(sost);
	std::vector<int> abs_rel = abs(rel);
	int k = 1;
	while (std::count(abs_sost.begin(), abs_sost.end(), k) || std::count(abs_rel.begin(), abs_rel.end(), k) || k==abs(e))
		k++;
	std::replace(rel.begin(), rel.end(), e, k);
	std::replace(rel.begin(), rel.end(), -e, -k);
	int_iterator r = rel.begin();
	while ((r = std::find(rel.begin(), rel.end(), k))!=rel.end())
		rel.insert(rel.erase(r), sost.begin(), sost.end());
	while ((r = std::find(rel.begin(), rel.end(), -k))!=rel.end())
		rel.insert(rel.erase(r), sost_inversed.begin(), sost_inversed.end());	
	cancella_coppie_opposte();
}

//se la relazione e' del tipo a^p per p>=0, restituisce p.
//Altrimenti restituisce -1
int relazione::ciclo() const {
  if (!lung()) return 0;
	if (std::count_if(rel.begin(), rel.end(), bind2nd(std::not_equal_to<int>(), rel.front())))
		return -1;
	return lung();
}

//se esiste un generatore che compare una sola volta, restituisce
//la posizione di questo generatore (partendo da 1 !)
//altrimenti restituisce 0
int relazione::esplicita() const {
	std::vector<int> abs_rel = abs(rel);
  for (int i=0; i<lung(); i++) 
		if (std::count(abs_rel.begin(), abs_rel.end(), abs_rel[i]) == 1)
			return i+1;
  return 0;
}

//restituisce il numero di occorrenze di 'e' e di '-e' nella relazione
int relazione::count(int e) const {
	return std::count(rel.begin(), rel.end(), e) + std::count(rel.begin(), rel.end(), -e);
}

//restituisce il numero di lettere nella relazione
int relazione::lung() const {
	return rel.size();
}

//restituisce la quantita' totale di lettere contenute nelle relazioni
int gruppo::lung() const {
	int ris=0;
	for (relazione_const_iterator r = relazioni.begin(); r!=relazioni.end(); r++)
		ris+=r->lung();
	return ris;
}

//dice se la relazione e' del tipo aba^(-1)b^(-1) con a!=b
bool relazione::commutatore() const {
	if (lung()!=4) return FALSE;
	if (rel[0]==rel[1]) return FALSE;
	return (rel[0]==-rel[2] && rel[1]==-rel[3]);
}

//cambia ogni occorrenza del generatore 'a' in 'ab' nelle relazioni
//funziona solo se abs(a)!=abs(b)
gruppo& gruppo::cambia_generatore(int a, int b) {
	if (abs(a) == abs(b)) return *this;
	std::vector<int> sost(2);
	sost[0] = a; sost[1] = b;
	for (relazione_iterator r = relazioni.begin(); r!=relazioni.end(); r++)
		r->sostituisci(a, sost);
	return *this;
}

//scambia i generatori 'a' e 'b' (funziona anche con b=-a)
void relazione::permuta(int a, int b) {
	std::replace(rel.begin(), rel.end(), a, 0);
	std::replace(rel.begin(), rel.end(), b, a);
	std::replace(rel.begin(), rel.end(), 0, b);		
	if (a!=b) {
		std::replace(rel.begin(), rel.end(), -a, 0);
		std::replace(rel.begin(), rel.end(), -b, -a);
		std::replace(rel.begin(), rel.end(), 0, -b);		
	}
}

//scambia i generatori 'a' e 'b' (funziona anche con b=-a)
gruppo& gruppo::permuta(int a, int b) {
	for (relazione_iterator r=relazioni.begin(); r!=relazioni.end(); r++)
		r->permuta(a,b);
	return *this;
}

//se num_lettere=0:
//  prende la prima relazione vuota e la elimina
//se num_lettere!=0:
//  prende la prima relazione esplicita (i.e. con una lettera x non ripetuta)
//  con lung() = num_lettere se num_letere>0, 
//  con lung() qualsiasi se num_lettere = -1,
//  elimina la relazione trovata,
//  sostituisce la x in tutte le altre relazioni (se lung()>0)
//Restituisce TRUE se trova una tale relazione, FALSE altrimenti.
bool gruppo::semplifica_con_sostituzioni(int num_lettere) {
	relazione_iterator r = relazioni.begin();
	while (r!=relazioni.end()) {
		if (num_lettere == 0 && !r->lung()) {
			relazioni.erase(r); 
			return TRUE;
		}				
		if (num_lettere > 0 && r->lung() == num_lettere)
			if (r->esplicita()) break;
		if (num_lettere == -1)
			if (r->esplicita()) break;
		r++;
	}
	if (r!=relazioni.end()) {
		if (num_lettere!=0) {
			int rid = r->esplicita()-1;
			std::vector<int> sost = r->get_rel();
			rotate(sost.begin(), sost.begin()+rid, sost.end());
			int e = -sost.front();
			sost.erase(sost.begin());
			for (relazione_iterator corr = relazioni.begin(); corr!=relazioni.end(); corr++)
				corr->sostituisci(e, sost);
			for (int i=e; i<=generatori; i++)
				permuta(i, i+1);
			generatori--;
		}
		relazioni.erase(r);
		return TRUE;
	}
	return FALSE;
}

//semplifica usando le sostituzioni, finche' esistono relazioni esplicite
//(cioe' contenenti una lettera una volta sola)
//restituisce TRUE se viene fatta almeno una sostituzione, FALSE altrimenti
bool gruppo::semplifica_con_sostituzioni() {
	bool ris = FALSE;
	for (int i=0; i<2;)
		if (semplifica_con_sostituzioni(i)) {
			i=0; ris = TRUE;
		}
		else i++;
	if (semplifica_con_sostituzioni(-1)) {
		semplifica_con_sostituzioni();
		ris = TRUE;
	}
	return ris;
}

//abbina cicli con la stessa lettera, ad esempio sostituisce a^12, a^8 con a^4
//restituisce TRUE se ci sono due cicli a^n, a^m, FALSE altrimenti
bool gruppo::semplifica_abbinando_cicli() {
	bool ris = FALSE;
	for (relazione_iterator r=relazioni.begin(); r!=relazioni.end(); r++)
		if (r->ciclo()>0) {
			int a = abs(r->get_rel(0));
			relazione_iterator r2=r; r2++;
			while (r2!=relazioni.end())
				if (r2->ciclo()>0 && abs(r2->get_rel(0))==a) {
					int M = MCD(r->lung(),r2->lung());
					for (int i=r->lung(); i<M; i++)
						r->pop_back();
					r2 = relazioni.erase(r2);
					ris = TRUE;
				}
				else r2++;
		}
	return ris;
}

//cerca di trovare una coppia a,b di generatori (con segno) tali che
//sostituendo a con a*b in ogni relazione (e poi semplificando con
//sostituzioni) diminuisce i generatori o la lung(). 
//Se trova la coppia a,b effettua la sostituzione e restituisce TRUE,
//restituisce FALSE altrimenti
bool gruppo::semplifica_cambiando_generatori() {
	for (int a=1; a<=generatori; a++)
		for (int b=1; b<=generatori; b++)
			if (a!=b) for (int segno=-1; segno<=1; segno+=2) {
				int lung_prec = lung();
				int generatori_prec = generatori;
				cambia_generatore(a,segno*b);
				semplifica_con_sostituzioni();
				if (generatori < generatori_prec || lung() < lung_prec)
					return TRUE;
				cambia_generatore(a,-segno*b);
				semplifica_con_sostituzioni();
			}
	return FALSE;
}

//usa 'semplifica_con_sostituzioni' e 'semplifica_abbinando_cicli'
//finche' hanno un effetto.
gruppo& gruppo::semplifica() {
	do
		do {
			while (semplifica_con_sostituzioni()) {}
		} while (semplifica_abbinando_cicli());
	while (semplifica_cambiando_generatori());
	return *this;
}

//mette in forma standard
gruppo& gruppo::forma_standard() {
	int sz = num_relazioni();
	for_each(relazioni.begin(), relazioni.end(), std::mem_fun_ref(&relazione::forma_standard));
	return *this;
}

//mette in forma standard
void relazione::forma_standard() {
  //inizia ogni componente con 'a','b' ... (e non con '-a', '-b')
	if (rel[0]<0) {
		inverti();
		ruota(lung()-1);
	}
}

//restituisce il numero di relazioni
int gruppo::num_relazioni() const {
	return relazioni.size();
}

//dice se i relatori sono cicli e commutatori
bool gruppo::forma_abeliana() const {
  for (relazione_const_iterator r = relazioni.begin(); r!=relazioni.end(); r++)
    if (r->ciclo()<0 && !r->commutatore()) return FALSE;
  return TRUE;
}

//dice se le lettere 'e' e '-e' non compaiono in nessun relatore
bool gruppo::libero (int e) const {
  for (relazione_const_iterator r=relazioni.begin(); r!=relazioni.end(); r++)
		if (r->count(e)) return FALSE;
  return TRUE;
}

//mette il gruppo abeliano in forma canonica, ad esempio Z24 x Z diventa Z x Z3 x Z8
//(sono ordinati gli Zp secondo p)
void gruppo_abeliano::forma_canonica() {
	int size_iniziale = p.size();
  for (int j=0; j<size_iniziale; j++)
    for (int k=2; k<=p[j]; k++)
      if (!(p[j] % k) && phi(k)==k-1) {
				int potenza_di_k=1;
				while (!(p[j] % k)) {
					potenza_di_k*=k; p[j]/=k;
				}
				if (p[j]!=1)
					p.push_back(potenza_di_k);
				else p[j] = potenza_di_k;
      }
	sort(p.begin(), p.end());
}

//scrive la presentazione su f
void gruppo::scrivi (std::ostream& f) const {
	char v[2];
	f << "<";
	for (int i=1; i<generatori; i++) {
		lettera(i,v); f << v << ", ";
	}
	if (generatori) {
		lettera(generatori,v); f << v << "|";
	}
	for (relazione_const_iterator r=relazioni.begin(); r!=relazioni.end(); r++) {
		if (r!=relazioni.begin()) f << ", ";
		int esp=1;
		for (int i=0; i<r->lung(); i++) {
			if (i==r->lung()-1 || r->get_rel(i) != r->get_rel(i+1)) {
				lettera(abs(r->get_rel(i)),v); f << v;
				if (r->get_rel(i)<0) esp*=-1;
				if (esp!=1) f << "^(" << esp << ")";
				esp=1;
			}
			else esp++;
		}
	}
	f << ">\n";
}

//scrive il gruppo abeliano su f
void gruppo_abeliano::scrivi (std::ostream& f) const {
  char w[6];
  if (p.empty()) f << "(e)";
  for (int i=0; i<p.size(); i++) {
    f << "Z";
    if (p[i]) {
      f << "_(" << p[i] << ")";
    }
    if (i+1<p.size()) f << "+";
  }
	f << "\n";
}

//dice se due gruppi abeliani sono uguali.
bool operator==(const gruppo_abeliano& g1, const gruppo_abeliano& g2) {
	gruppo_abeliano h1 = g1;
	gruppo_abeliano h2 = g2;
	h1.forma_canonica(); h2.forma_canonica();
	return (h1.p == h2.p);
}
