//per collegare il tutto: gcc -w -g -lstdc++ og*.o   -w no warning, -g debugging infos
//per far partire: exec -name a.out

#if !defined (OGLIST_CPP)
#define OGLIST_CPP

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <malloc.h>
#include <iostream>
#include <fstream>
#include <list>
#include <vector>
#include <algorithm>

#include "defs.h"
#include "math.h"
#include "ograph.h"
#include "list.h"
#include "io_base.h"

typedef std::list<ografo>::const_iterator ografo_const_iterator;
typedef std::list<ografo>::iterator ografo_iterator;

extern bool mostra_contatore;

//restituisce una lista di ografi come segue:
//se param == 1: fa una MP su ogni spigolo
//se param == 2: fa una lune(-1) su ogni bigono
//se param == 3: fa una MP^(-1) su ogni triangolo
//se param == 4: fa una mossa-4 su ogni quadrato e colora la nuova faccia con -1
//   se param2 == 1: non fa mossa-4 sulle facce con colore!=0
//   attenzione: non restituisce l'ografo iniziale!
template<class T> std::list<T> disc_replacement(const T& o, int param, int param2) {
  std::list<T> ls;
  const int l=o.lung();
  if (param == 1) {
    for (int i=0; i<l; i++)
      for (int j=0; j<4; j++)
        if (o.prima_occorrenza_spigolo(i,j) && !o.falso(i,j) && o.get_v(i,j)!=i) {
          T o2(o);
          ls.push_back(o2.MP(i,j));
        }
  }
  if (param == 2 || param == 3 || param == 4) {
    int contatore = 1; int tipo=0;
		T o2;
		while ((o2=o).disc_replacement(param, contatore, tipo, param2)) {
			ls.push_back(o2);
			if (param <= 3) contatore++;
			if (param == 4) {

				//colora il quadrato trasformato con -1
				const int primo = o2.lung()-3;
				for (int i=0; i<o2.get_ff().size(); i++) {
					const faccia& f2 = o2.get_ff()[i];
					for (int k=0; k<f2.lati(); k++)
						if (f2.get_passaggio(k) == primo && (f2.get_uscita(k) == 1 && f2.get_entrata(k)==2 || 
																								 f2.get_uscita(k) == 2 && f2.get_entrata(k)==1))
							f2.set_colore(-1);
						else f2.set_colore(0);
				}
				
				if (tipo) contatore++;
				tipo = 1-tipo;
			}
		}
  }
  return ls;
}

//se param == 1,2,3: restituisce una lista di tutti i poliedri ottenibili
//rispettivamente con MP, lune^(-1) e MP^(-1)
//  se param2 == 1 (DEFAULT) mette gli ografi in forma_minima ed 
//  elimina i duplicati della lista ottenuta, 
//  altrimenti la restituisce come e'.
//
//se param == 4: restituisce una lista di ografi in forma_minima, 
//senza duplicati (se lo e' ol!), 
//di tutti i poliedri ottenibili da ol con successive mosse-4, incluso 'ol'.
//  se param2 >= 0: si ferma al livello 'param2' (compreso, partendo da 1, 
//  quindi se param2==0 restituisce una lista uguale a ol)
//  se param2 == -1: si ferma solo quando ha trovato tutti i poliedri ottenibili
template<class T> std::list<T> disc_replacement(const std::list<T>& ol, int param, int param2) {
  std::list<T> ls = std::list<T>();
  if (ol.empty()) return ls;
  const int l = ol.front().lung(); 
  typedef typename std::list<T>::const_iterator T_const_iterator; 
  if (param == 1 || param == 2 || param == 3) {
    int contatore = 1;
    const int lg = ol.size();
    for (T_const_iterator o = ol.begin(); o!=ol.end(); o++) {
      if (mostra_contatore) scrivi_contatore("Mosse MP : ", l, contatore++, lg);
      std::list<T> ls2 = disc_replacement(*o, param, 0);
      ls.splice(ls.end(), ls2);
    }
    if (param2 == 1) elimina_poliedri_doppi(ls);
    if (mostra_contatore) {
      cancella_riga(60); scrivi_info(ls);
    }
  }
  if (param == 4) {
    int livello=0;
    int quanti_prima=0;
    std::vector<T> ls_vector;
    for (T_const_iterator o = ol.begin(); o!=ol.end(); o++) {
      ls_vector.push_back(*o);
      const T o2 = ls_vector.back();
			for (int i = 0; i < o2.get_ff().size(); i++) o2.get_ff()[i].set_colore(0);
    }      
		int quanti = ls_vector.size();
    while ((param2==-1 || (livello++)<param2) && quanti>quanti_prima) {
      int lg = ls_vector.size();
			std::list<T> nuova;
      for (int i = quanti_prima; i < quanti; i++) {
        if (mostra_contatore) scrivi_contatore("Mosse 4: ", l, i, lg);
				std::list<T> dr = disc_replacement(ls_vector[i], 4, 1);
        nuova.splice(nuova.end(), dr);
      }
      ls_vector.insert(ls_vector.end(), nuova.begin(), nuova.end());
      elimina_poliedri_doppi(ls_vector, quanti);
      quanti_prima = quanti;
      quanti = ls_vector.size();
      if (mostra_contatore) {
        cancella_riga(60);
        scrivi_info(ls_vector);
        }
    }
    ls = std::list<T>(ls_vector.begin(), ls_vector.end());      
  }
	return ls;
}

//se due elementi di ls sono uguali, elimina il secondo. Va avanti cosi' finche'
//non sono tutti distinti.
//Usa il fatto che i primi 'num_gia_fatti' (partendo da 1, default = 0) sono gia' in forma
//minima e distinti fra loro 
template<class T> void elimina_poliedri_doppi(T& ls, int num_gia_fatti) {
  if (ls.empty()) return;
  int l = ls.front().lung(); int lg = ls.size();
  typedef typename T::iterator T_iterator; 
  typedef typename T::const_iterator T_const_iterator; 
  T_iterator o = ls.begin();
  for (int contatore=0; contatore<lg; contatore++, o++) if (contatore>=num_gia_fatti) {
    if (mostra_contatore) scrivi_contatore("Set forma minima : ", l, contatore, lg);
    o->set_forma_minima();
  }
  if (mostra_contatore) cancella_riga(60);
  T_iterator ultimo_gia_fatto = ls.begin();
  for (int t=0; t<num_gia_fatti-1; t++)
    ultimo_gia_fatto++;
  for (T_const_iterator o = ls.begin(); o!=ls.end();o++) {
    T_iterator o2 = ultimo_gia_fatto; o2++;
    while (o2!=ls.end()) (*o == *o2) ? o2 = ls.erase(o2) : o2++;
    if (o == ultimo_gia_fatto)
      ultimo_gia_fatto++;
  }
}

template<class T> void elimina_varieta_doppie(T& ls, int profondita) {
	if (ls.empty()) return;
  int l = ls.front().lung(); int lg = ls.size();
	T ris;
  typedef typename T::iterator T_iterator; 
  typedef typename T::const_iterator T_const_iterator; 
  for (T_iterator o = ls.begin(); o!=ls.end(); o++)
		o->set_forma_minima();
	int contatore = 0;
  for (T_iterator o = ls.begin(); o!=ls.end(); contatore++, o++) {
    //if (mostra_contatore) scrivi_contatore("Elimina varieta doppie : ", l, contatore, lg);
		T cugine = T(1,*o);
		cugine = disc_replacement(cugine, 4);
		for (int t=0; t<profondita; t++) {
			cugine = disc_replacement(cugine, 1);
			cugine = disc_replacement(cugine, 4);
		}
		for (int t=0; t<profondita; t++) {
			cugine = disc_replacement(cugine, 3);
			cugine = disc_replacement(cugine, 4);
		}
		for (T_iterator o2 = cugine.begin(); o2!=cugine.end(); o2++)
			o2->set_forma_minima();
		bool nuova = TRUE;
		for (T_iterator o2 = cugine.begin(); o2!=cugine.end(); o2++)
			for (T_iterator o3 = ls.begin(); o3!=o; o3++)
				if (*o2 == *o3)
					nuova = FALSE;
		if (nuova)
			ris.push_back(*o);
	}
	if (mostra_contatore) std::cout << "\n";
	ls = ris;
}

//da usare con profondita<=2 (meglio <=1) per evitare che faccia due lune di fila
template<class T> void elimina_non_minimali(T& ls, int profondita_su, bool usa2, int profondita4) {
	if (ls.empty()) return;
	T ris;
  typedef typename T::iterator T_iterator; 
  typedef typename T::const_iterator T_const_iterator; 
	int contatore = 0;
  for (T_iterator o = ls.begin(); o!=ls.end(); contatore++, o++) {
    std::cout << contatore << std::endl;
		T min = parenti_minimali(*o, profondita_su, usa2, profondita4, 1 + profondita_su);
		if (min.front().lung() == o->lung()) ris.push_back(*o);
	}
	ls = ris;
}	

//scrive informazioni sull'olista.
template<class T> void scrivi_info(const T& ol) {
  if (!ol.size())
    std::cout << "Lista vuota\n";
  else {
    std::cout << "Numero ografi con ";
    int l = ol.size() ? ol.front().lung():0;
    std::cout << l << " vertici: " << ol.size() << std::endl;
  }
}

//carica dal file 'file'.
template<class T> std::list<T> carica(std::ifstream& file, int tipo_formato) {
  std::list<T> ls;
  char w[2];
  for (int i=1; file.read(w,1); i++) {
    file.seekg(-1, std::ios::cur);
		T o = T(file, tipo_formato);
    ls.push_back(o);
  }
  return ls;
}

//carica dal file avente nome 'nome_file'
template<class T> std::list<T> carica(char* nome_file, int tipo_formato) {
  std::ifstream f(nome_file, std::ios::in);
	return carica<T>(f, tipo_formato);
}

//salva sul file 'file'.
template<class T> void salva(const std::list<T>& ls, std::ofstream& file, int tipo_formato) {
  typename std::list<T>::const_iterator i; 
	for (i=ls.begin(); i!=ls.end(); i++)
		i->salva(file, tipo_formato);
}

//salva sul file avente nome 'nome_file' con formato 'tipo_formato'
template<class T> void salva(const std::list<T>& ls, char* nome_file, int tipo_formato, bool append) {
  std::ofstream f(nome_file, append ? std::ios::app : std::ios::out);
	salva<T>(ls, f, tipo_formato);
}

//scrivi. Funziona se T e' derivata da 'ografo'.
template<class T> void scrivi(const std::list<T>& ls, std::ostream& f, int tipo_formato) {
	typename std::list<T>::const_iterator i;
	for (i=ls.begin(); i!=ls.end(); i++)
		scrivi(*i, f, tipo_formato);
}

template<class T> std::list<T> parenti_minimali(const T& o, int profondita_su, bool usa2, int profondita4, int profondita_giu) {
	std::list<T> ol = std::list<T>(1,o);
	//ol.front().set_forma_minima();  Serve ?
	return parenti_minimali(ol, profondita_su, usa2, profondita4, profondita_giu);
}

template<class T> std::list<T> parenti_minimali(const std::list<T>& ol, int profondita_su, bool usa2, int profondita4, int profondita_giu) {
	if (profondita_giu == 0) return ol;
  if (profondita_su) {
    std::list<T> ol2 = ol;
    for (int i=0; i<profondita_su; i++)
      ol2 = disc_replacement(ol2,1,1);
    return parenti_minimali(ol2, 0, usa2, profondita4, profondita_giu);
  }
  std::list<T> ol2 = disc_replacement(ol, 4, profondita4);
  std::list<T> ol3;
	if (usa2) ol3 = disc_replacement(ol2, 2, 1);
  if (ol3.empty()) ol3 = disc_replacement(ol2, 3, 1);
  if (!ol3.empty()) return parenti_minimali(ol3, 0, usa2, profondita4, profondita_giu - 1);
  else return ol2;
}

#endif
