#include <stdlib.h>
#include <string>
#include <math.h>
#include <ctype.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <list>
#include <algorithm>

#include "defs.h"
#include "math.h"
#include "templates.cpp"
#include "string.h"
#include "ograph.h"
#include "io.h"
#include "surfaces.h"

int svincolo[4][3]={{1,2,3},{2,3,0},{3,0,1},{0,1,2}};

extern const int Aquattro[12][4];
extern const int Aquattroinv[12][4];
extern bool inglese;

//costruttore standard. Mette vicino[t]=-1
overtice::overtice():ivertice() {
  for(int t=0; t<4; t++) tor[t]=0;
}

//costruttore di copia
ografo::ografo (const ografo& o):ff(o.ff) {
	for (int i=0; i<o.lung(); i++)
		v.push_back(o[i]->clone());
}

//carica ografo dal file in formato Ographs se tipo_formato=0 (DEFAULT) 
//con formato SnapPea se tipo_formato=1
//e con formato SnapPea ristretto (senza informazioni sul bordo o angoli) se tipo_formato=2
ografo::ografo (std::istream& file, int tipo_formato): ff() {
  if (tipo_formato==0) {
    char w[200];
    file.getline(w, 100);
    const int l=ctoi(&w[11],2);
    for (int i=0; i<l; i++) v.push_back(new overtice());
    for (int i=0; i<l; i++) {
      file.getline(w,100);
      for (int j=0; j<4; j++) {
        int i2 = ctoi(&w[9+2*j],2);
        int j2 = ctoi(&w[21+2*j],2);
        int tr = ctoi(&w[33+2*j],2);
				set_col(i,j,i2,j2,tr);
      }
    }
    file.getline(w,100);  //prende la linea di chiusura '--------'
  }
  if (tipo_formato == 1 || tipo_formato == 2) {
    char w[100];
    int num_riga=0;
    if (tipo_formato == 1) while (file.getline(w, 100) && num_riga<8) num_riga++;
    while (file.getline(w, 100) && w[1]==' ') num_riga++;
    int l=ctoi(w,2);
    for (int i=0; i<l; i++) v.push_back(new overtice());
    for (int i=0; i<l; i++) {
      file.getline(w,100);
      for (int j=0; j<4; j++)
        v[i]->set_vicino(j,ctoi(&w[5*j],4));
      file.getline(w,100);
      for (int usc=0; usc<4; usc++) {
        int perm[4];
        for (int j=0; j<4; j++) 
          perm[j]=ctoi(&w[5*usc+1+j],1);
        v[i]->set_col(usc,perm[usc]);
        int x=perm[mod(usc+1,4)]-perm[usc]-1;
        x=mod(x,4);
        torsione tr;
        if (mod(usc,2)==0 && mod(perm[usc],2)==0) tr = mod(2-x,3);
        if (mod(usc,2)==1 && mod(perm[usc],2)==1) tr = mod(x+1,3);
        if (mod(usc,2)==0 && mod(perm[usc],2)==1) tr = mod(x,3);
        if (mod(usc,2)==1 && mod(perm[usc],2)==0) tr = mod(-x,3);
        v[i]->set_tor(usc,tr);
      }
      if (tipo_formato == 1) for (int i=0; i<7; i++)
        file.getline(w,100);
    }
  }
	set_facce();
}

//costruisce ografo da igrafo, mettendo la torsione tor[i] sullo spigolo i-esimo
//(partendo da zero)
ografo::ografo(const igrafo& ig, const std::vector<torsione>& tr): igrafo(), ff() {
	const int l = ig.lung();
	for (int i=0; i<l; i++)
		v.push_back(new overtice());
	int spigolo = 0;
	for (int i=0; i<l; i++)
		for (int j=0; j<4; j++)
			if (get_col(i,j) == -1)
				set_col(i, j, ig.get_v(i,j), ig.get_col(i,j), tr[spigolo++]);
	set_facce();
}

//dice se i dati dell'ografo corrispondono effettivamente ad un ografo
bool ografo::vero() const {
  for (int i=0; i<v.size(); i++)
    for (int j=0; j<4; j++) {
      int i2 = get_v(i,j);
      int j2 = get_col(i,j);
      if (get_v(i2,j2)!=i) return FALSE;
      if (get_tor(i,j)!=get_tor(i2,j2)) return FALSE;
    }
  return TRUE;
}

//costruisce le facce
void ografo::set_facce() {
  ff = facce(*this);
}

//ritorna il numero di 1-scheletro nella lista schel#.dat (con #=lung()+1)
//che ha scheletro uguale a quello dell'ografo (partendo da 1)
//ritorna 0 se non lo trova
int ografo::num_1scheletro() const {
  grafo g = grafo(*this);
  g.set_forma_minima();
  int l=lung();
  char z[14] = "schel"; strcat(z,l); strcat (z, ".dat");
  std::ifstream file(z, std::ios::in);
  for (int i=1; file.read(z,1); i++) {
    file.seekg(-1,std::ios::cur);
    grafo g2 = grafo (file);
    if (g==g2) return i;
  }
  return 0;
}   

//restituisce gli spigoli falsi (cioe' con tor==3)
std::vector<int> ografo::spigoli_falsi() const {
  std::vector<int> ls(0);
  int l=lung();
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++)
      if (prima_occorrenza_spigolo(i,j) && falso(i,j))
        ls.push_back(v2spigolo(i,j));
  return ls;
}

//restituisce il numero di eulero
int ografo::eulero() const {
  return ff.size()-lung();
}

//dice se e' lo scheletro di una coppia 
//ATTENZIONE: in verita' fa due controlli:
//1) che ogni faccia sia adiacente ad un solo spigolo falso al piu'
//2) che la caratteristica sia giusta
bool ografo::scheletro() const {
	for (int i=0; i<ff.size(); i++) {
		const faccia& f = ff[i];
		int lt = f.lati();
		int spig1 = v2spigolo(f.get_passaggio(0), f.get_entrata(0));
		int spig2 = v2spigolo(f.get_passaggio(lt-1), f.get_uscita(lt-1));
		if (spig1!=spig2) return FALSE;
	}
  int qsf = spigoli_falsi().size();
	if (eulero() != 1 + qsf) return FALSE;
  return TRUE;
}

//pseudo_minimale = non contiene non-fake loops con l<=3 
//ATTENZIONE: ora per l=3 cerca solo facce embedded con 3 lati, e anche per l=2 non cerca
//sugli spigoli finti e cerca solo loops contenuti in coppie di facce distinte
bool ografo::pseudo_minimale(std::string* ris) const {
  int l=lung();        //facce embedded con l<=3
  int cc = ff.size();
  for (int i=0; i<cc; i++)
    if (chiusa(ff[i]) && ff[i].embedded() && ff[i].lati()<=3) {
      if (ris) {
        ris->append (inglese ? "No, it contains an embedded face of length " : 
										 "No, contiene faccia embedded di lunghezza "); 
				char temp[4];
        ris->append (gcvt(ff[i].lati(),3,temp)); ris->append("\n"); 
      }
      return FALSE;
    }
  if (scheletro()) {
    std::vector<teta> th = get_teta(*this) ;
    for (int i=0; i<2*l; i++) {
      if (th[i].counterpass()) {
        if (ris) ris->append(inglese ? "No, it contains a length-1 loop \n" : 
														 "No, contiene loop di lunghezza 1 \n");
        return FALSE;
      }
      if(th[i].triple_counterpass()) {
        if (ris) ris->append(inglese ? "No, it contains a triple counterpass\n" :
														 "No, contiene triple counterpass\n"); 
        return FALSE;
      }
    }
    for (int i=0; i<2*l; i++) 
      for (int i2=i+1; i2<2*l; i2++)
        if (short_loop(th[i],th[i2])) {
          if (ris)
            ris->append(inglese ? "No, it contains a non-fake loop of length 2\n" :
												"No, contiene non-fake loop di lunghezza 2\n");
          return FALSE;
        }
  }
  if (ris) ris->append(inglese ? "Yes\n" : "Si\n");
  return TRUE;
}

//mossa MP su spigolo(i,j)
ografo& ografo::MP(int i, int j) {
  int l=lung();
  if (j>4 || i>l) return *this;
  if (falso(i,j)) return *this;
  int i2=get_v(i,j);
  if (i==i2) return *this;
  if (mod(j,2)) {
    C(i,1); j--;
  }
  if (j==0) C(i,2);
  if (mod(get_col(i,2),2)) C(i2,1);
  if (get_col(i,2)==2) C(i2,2);
  while (get_tor(i,2)!=torsione(0)) C(i,8);
  
	aggiungi_vertici(3);
  int nuovi[4][2] = {{2,0},{0,3},{-1,-1},{1,3}};
  int nuovi2[4][3] = {{-1,-1,-1},{0,0,1},{2,2,0},{1,2,2}};
  for (int t=0; t<4; t++) if (t!=2)
    set_col(l+nuovi[t][0], nuovi[t][1], get_v(i,t), get_col(i, t), get_tor(i,t));
  for (int t=0; t<4; t++) if (t!=0)
    set_col(l+nuovi2[t][0], nuovi2[t][1], get_v(i2,t), get_col(i2, t), get_tor(i2,t)+=nuovi2[t][2]);
	set_col(l,2,l+2,1,2);
  set_col(l,1,l+1,1,0);
  set_col(l+1,0,l+2,3,1);
  rimuovi(i);
  rimuovi(i<i2 ? i2-1 : i2);
  set_facce();
  return *this;
}

//mette in forma standard la faccia ff[num_faccia] (parte quindi da zero),
//se e' chiusa ed embedded.
//forma standard: (entrata[t], uscita[t]) = (1,0) per ogni t
//restituisce TRUE se era chiusa embedded. Altrimenti restituisce FALSE
//e non fa nulla
bool ografo::set_faccia_su_01 (int num_faccia) {
	if (num_faccia < 0 || num_faccia >= ff.size()) return FALSE;
  faccia& f = ff[num_faccia];
	if (!chiusa(f) || !f.embedded()) return FALSE;
  const int l=lung();
  const int lati = f.lati();
  for (int j=0; j<lati; j++) {
    int vert = f.get_passaggio(j);
    int usc = f.get_uscita(j);
    if (mod(usc,2)) C(vert,1);
    if (usc>=2) C(vert,2);
    while (f.get_entrata(j)!=1)
      C(vert,4);
  }
  return TRUE;
}

//buca la faccia numero 'num_faccia' (partendo da zero). 
//Funziona solo se f e' chiusa ed embedded
//restituisce TRUE in questo caso, FALSE altrimenti
//ATTENZIONE, controllare la definizione di 'tors': ho messo '+1'
bool ografo::buca(int num_faccia) { 
  if (!set_faccia_su_01(num_faccia)) return FALSE;
	faccia f = ff[num_faccia];
  const int lati = f.lati();
  for (int i=0; i<lati; i++) {
    int vicini[2]; int ent[2];
    for (int t=0; t<2; t++) {
      vicini[t]=get_v(f.get_passaggio(i),t+2); 
      ent[t]=get_col(f.get_passaggio(i),t+2);
    }
    torsione tors = get_tor(vicini[0],ent[0]) + get_tor(vicini[1],ent[1]) + 1;
    set_col(vicini[0],ent[0], vicini[1], ent[1], tors);
  }
  std::vector<int> z = f.get_passaggi();
  sort(z.begin(), z.end());
  for (int i=lati-1; i>=0; i--) 
    rimuovi(z[i]);
  set_facce();
	return TRUE;
}

//prende lo spigolo falso num-esimo (partendo da 1) e collassa
//il triodo corrispondente
//funziona solo se le tre facce adiacenti al triodo hanno incidenza totale
//su ogni spigolo al piu' uno (che e' garantito per i mattoni con c>3)
bool ografo::collassa_triodo_di_bordo(int num) {
	std::vector<int> spigoli = spigoli_falsi();
	if (num<=0 || num>spigoli.size()) return FALSE;
	int spigolo = spigoli[num-1];

	//trasforma ogni vertice in modo che       f     f      f1                  
	//ciascuna faccia f,g,h sia adiacente     f*n   f*g   f1*f2
	//in uno di questi modi al vertice         m     g      f2  (m,n!=f,g,h)
	//(eccetto i due vertici adiacenti a spigolo)
	std::vector<int> facce_adiacenti = std::vector<int>();
	for (int i=0; i<ff.size(); i++)
		if (v2spigolo(ff[i].get_passaggio(0), ff[i].get_entrata(0)) == spigolo) {
			facce_adiacenti.push_back(i);
			for (int j=1; j<ff[i].lati()-1; j++) {
				int vert = ff[i].get_passaggio(j);
				int usc = ff[i].get_uscita(j);
				if (mod(usc,2)) C(vert,1);
				if (usc>=2) C(vert,2);
				while (ff[i].get_entrata(j)!=1)
					C(vert,4);
			}
		}

	//elimina le facce f,g,h
	std::vector<int> vertici_toccati = std::vector<int>();
	for (int nf=0; nf<facce_adiacenti.size(); nf++) {
		faccia& f = ff[facce_adiacenti[nf]];
		for (int i=1; i<f.lati()-1; i++) if (f.get_entrata(i)==1) {
			int vicini[2]; int ent[2];
			for (int t=0; t<2; t++) {
				vicini[t]=get_v(f.get_passaggio(i),t+2); 
				ent[t]=get_col(f.get_passaggio(i),t+2);
			}
			torsione tors = get_tor(vicini[0],ent[0]) + get_tor(vicini[1],ent[1]) + 1;
			set_col(vicini[0],ent[0], vicini[1], ent[1], tors);
			vertici_toccati.push_back(f.get_passaggio(i));
		}
		vertici_toccati.push_back(f.get_passaggio(0));
		vertici_toccati.push_back(f.get_passaggio(f.lati()-1));
	}
	elimina_duplicati(vertici_toccati);
  for (int i=vertici_toccati.size()-1; i>=0; i--) 
    rimuovi(vertici_toccati[i]);
  set_facce();	
}

//se lati == 2, 3 o 4:
//  fa disc_replacement sulla faccia numero 'num_faccia' (partendo da 1) 
//  tra quelle aventi lati = 'lati' (che sono 3 o 4)
//  restituisce 0 se non c'e' questa faccia, un numero maggiore di zero altrimenti
//  Se 'lati'==4: 
//    tipo == 0 oppure tipo == 1 determina quale delle due mosse fare.
//    Se ci riesce: restituisce il numero della nuova faccia creata (partendo da 1)
//se 'solo_colore_zero' == TRUE (default = FALSE), conta solo le facce con colore zero.
int ografo::disc_replacement (int lati, int num_faccia, bool tipo, bool solo_colore_zero) {
	if (lung()<=2) return FALSE;
  if (lati==2 || lati==3 || lati==4) {
		int contatore = 0;
		int sz = ff.size();
		for (int i=0; i<sz; i++) 
			if (chiusa(ff[i]) && ff[i].embedded() && ff[i].lati()==lati)
        if (!solo_colore_zero || ff[i].get_colore() == 0)
          if (++contatore == num_faccia) {
            return disc_replacement (i, tipo);
          }
		return 0;
	}
}

//Sia f = ff[num_faccia] (quindi parte da zero)
//se f.lati()==2 fa una lune(-1) (se crea componente anulare, la sostituisce con 2 dischi)
//se f.lati()==3 fa una MP^(-1)
//se f.lati()==4 fa una delle due mosse possibili, determinate da tipo==0 o tipo==1
//restituisce zero se non fa la mossa, un numero positivo altrimenti
//se f.lati()==4 e fa la mossa, restituisce il numero della nuova faccia (partendo da 1)
int ografo::disc_replacement (int num_faccia, bool tipo) {
  if (!set_faccia_su_01(num_faccia)) return FALSE;
	faccia f = ff[num_faccia];
  int l=lung();
  int h=0; vertice* pt=0;
  int lati = f.lati();
  std::vector<int> w(lati);  
	int y3[6][5] = {{0,3,0,0,2},{0,2,1,2,2},{1,3,0,1,1},{1,2,1,1,0},
								 {2,3,0,3,0},{2,2,1,3,1}};
  int y4[8][5] = {{0,3,0,0,0},{0,2,1,0,1},{1,3,0,3,2},{1,2,1,1,2},
                 {2,3,3,3,2},{2,2,2,1,2},{3,3,3,2,1},{3,2,2,2,0}};
  if (lati<2 || lati>4) return 0;
	switch (lati) {
    case 2:
      w[0]=f.get_passaggio(0); w[1]=get_v(w[0],0);
      for (int t=0; t<2; t++) {
        int vert1 = get_v(w[0], 2+t);
        int usc1 = get_col(w[0], 2+t);
        int vert2 = get_v(w[1], 2+t);
        int usc2 = get_col(w[1], 2+t);
        torsione tor = get_tor(w[0],2+t) + get_tor(w[1],2+t) + 1;
        set_col(vert1, usc1, vert2, usc2, tor);
      }
      break;
    case 3:
      w[0]=f.get_passaggio(0); w[1]=get_v(w[0],0); w[2]=get_v(w[1],0);
      aggiungi_vertici(2);
      for (int t=0; t<6; t++) {
        int vert = w[y3[t][0]]; int usc = y3[t][1];
        int vert1 = get_v(vert,usc); 
        int usc1 = get_col(vert,usc);
        int vert2 = l+y3[t][2]; 
        int usc2 = y3[t][3];
        torsione tor = get_tor(vert, usc)+=y3[t][4];
        set_col(vert1, usc1, vert2, usc2, tor);
      }
      set_col(l,2,l+1,0,0);
      break;
    case 4:
      aggiungi_vertici(4);
      w[0+tipo]=f.get_passaggio(0); for (int t=1; t<4; t++) w[(t+tipo) % 4]=get_v(w[t-1+tipo],0); 
      for (int t=0; t<8; t++) {
        int vert = w[y4[t][0]]; int usc = y4[t][1];
        int vert1 = get_v(vert,usc); 
        int usc1 = get_col(vert,usc);
        int vert2 = l+y4[t][2]; 
        int usc2 = y4[t][3];
        torsione tor = get_tor(vert, usc)+=y4[t][4];
        set_col(vert1, usc1, vert2, usc2, tor);
      }
      for (int t=0; t<4; t++)
        set_col(l+t, (1+t) % 4, l+((t+1)%4), (3+t) % 4,0);
  }
  sort(w.begin(), w.end());
  for (int k=lati-1; k>=0; k--)
    rimuovi(w[k]);
	set_facce();
  if (lati == 4) 
    for (int i=0; i<ff.size(); i++) if (ff[i].lati() == 4)
      for (int j=0; j<4; j++) {
        int vert = ff[i].get_passaggio(j); int usc = ff[i].get_uscita(j);
        int ent = ff[i].get_entrata(j);
        if (vert == l-4 && (usc == 1 || ent == 1))
          return i+1;
      }
  return 1;
}

//restituisce la terna di facce adiacenti allo spigolo CON SEGNO
//ATTENZIONE: le facce che ritorna sono numerate partendo da 1 (con segno!)
teta ografo::facce_su_spigolo(int vert, int usc) const {
  colora_progressivamente(ff);
  return teta(ff, vert, usc);
}

//Prende il num_spig-esimo spigolo (partendo da 1) falso, 
//che corrisponde ad un toro T nel bordo di M. Una
//base per H1(T) e' costruita (a meno di segno) chiamando (1,0),(0,1) e (1,1) 
//le 3 slope che si ottengono
//assegnando rispettivamente le torsioni 0,1,2 allo spigolo.
//(quest'ultima frase e' da verificare bene)
//restituisce TRUE se ce la fa, FALSE altrimenti
bool ografo::fill(int num_spig, int p, int q) {
  if (p<0) {p=-p; q=-q;}
  std::vector<int> spigoli = spigoli_falsi();
  if (num_spig>spigoli.size()) return FALSE;
  int spig = spigoli[num_spig-1];
  int vr, cl; spigolo2v(spig, vr, cl);  
  int tor, tor2;
  int cc = ff.size();
  if (!norma(p,q)) {
    if (!q) tor=0; if (!p) tor=1; if (p*q) tor=2;
    set_tor(vr,cl,tor);
    set_facce();
    teta th = facce_su_spigolo(vr, cl);
    if (!buca(abs(th.faccia_sola())-1)) {
			set_tor(vr,cl,3);
			set_facce();
			return FALSE;
		}
  }
	else {
    if (p<q) tor=0; if (p>q) tor=1; if (q<0) tor=2;
    set_tor(vr,cl,tor);
    set_facce();
    if (norma(p,q)==1) return TRUE;
    int l=lung();
    int vr2 = get_v(vr,cl); int cl2 = get_col(vr,cl);
    bool trovata=FALSE;
    int corsia, corsia2;
    for (int i=0; i<ff[i].lati(); i++) {
      int quanti=0;
      for (int k=0; k<ff[i].lati(); k++) if (ff[i].get_passaggio(k) == vr)
        if (ff[i].get_uscita(k)==cl || ff[i].get_entrata(k)==cl) quanti++;
      if (quanti==1) for (int k=0; k<ff[i].lati(); k++) {
        int usc = ff[i].get_uscita(k); int ent = ff[i].get_entrata(k);
        if (ff[i].get_passaggio(k) == vr) {
          if (usc==cl) corsia=mod(usc-ent-1,4);
          if (ent==cl) corsia=mod(ent-usc-1,4);
        }
        if (ff[i].get_passaggio(k) == vr2) {
          if (usc==cl2) corsia2=mod(usc-ent-1,4);
          if (ent==cl2) corsia2=mod(ent-usc-1,4);	    
        }
      }
    }
    if (mod(cl,2)) corsia = 2-corsia;
    if (mod(cl2,2)) corsia2 = 2-corsia2;
    int tor=corsia;
    int tor2=corsia2;
		aggiungi_vertici(1);
    set_col(l,0,l,1,3);
    set_col(l,2,vr,cl,tor);
    set_col(l,3,vr2,cl2,tor2);
    trasformazione trasf = trasformazione_banale();
    for (int i=0; i<l+1; i++) trasf.t[0][i] = mod(i+1,l+1);
    //problema qui:    
    trasforma(trasf);
    if (q>0 && p<q) fill(1,p,q-p); if (q>0 && p>q) fill(1,p-q,q); if (q<0) fill(1,-q,p);
 }
	return TRUE;
}

//mette tor=3 allo spigolo numero 'num' (partendo da 1) 
//tra quelli su cui poggia una mobius triplet.
//se num_spig==0 allora metter tor=3 su tutti gli spigoli che trova.
//ritorna TRUE se la tor di almeno uno spigolo e' cambiata
bool ografo::drill(int num) {
  std::vector<int> spigoli = spigoli_con_mobius_triplet();
  if (num<=0 || num>spigoli.size()) return FALSE;
  int spigolo = spigoli[num-1];
  int i,j;
  spigolo2v(spigolo,i,j);
  set_tor(i, j, 3);
  set_facce();
  return TRUE;
}

//attacca il toro di bordo allo spigolo falso i-esimo (partendo da 1)
//per adesso funziona solo se lo spigolo e' embedded
bool ografo::attacca_toro_di_bordo(int num) {
	std::vector<int> spigoli = spigoli_falsi();
	if (num<=0 || num>spigoli.size()) return FALSE;
	int spigolo = spigoli[num-1];
	int i,j;
	spigolo2v(spigolo,i,j);
	int i2 = get_v(i,j);
	int j2 = get_col(i,j);
	if (i==i2) return FALSE;
	//                                  |      |
	//mette lo spigolo nella forma     -i-(3)-i2-
	//                                  |      |
	if (j % 2) C(i,1);
	if (j<=1) C(i,2);
	if (j2 % 2) C(i2,1);
	if (j2>=2) C(i2,2);
	//                                  f      h
	//mette le tre facce nella forma   gi-(3)-i2g
	//                                  h      f
	bool ok = FALSE;
	while (!ok) {
		C(i2,4);
		for (int t=0; t<ff.size(); t++) 
			for (int caso=0; caso<=1; caso++) {
				faccia& f = ff[t];
				if (caso) f.inverti();
				if (f.get_passaggio(0) == i && f.get_entrata(0) == 2 && 
						f.get_uscita(0) == 0 && f.get_entrata(f.lati()-1) == 2)
					ok = TRUE;
			}
	}
	//crea i nuovi vertici e spigoli
	const int l = lung();
	aggiungi_vertici(2);
	set_col(i,2,l,0,0);
	set_col(i2,0,l+1,2,0);
	set_col(l,1,l+1,3,2);
	set_col(l,2,l+1,0,0);
	set_col(l,3,l+1,1,1);
	set_facce();
}

ografo& ografo::scheletro2spina() {
	if (!scheletro()) return *this;
	int quanti_spigoli_falsi = spigoli_falsi().size();
	for (int t=1; t<quanti_spigoli_falsi; t++)
		attacca_toro_di_bordo(1);
	collassa_triodo_di_bordo(1);
	return *this;
}

//ogni volta che tre facce distinte si incontrano su due spigoli distinti
//determinano un triodo. Questa funzione restituisce questi triodi, 
//nella forma di un vettore tr[][3]. 
//In tr[i] stanno le tre facce contenenti al triodo i-esimo (partendo da zero)
tensor2 ografo::triodi() const {
  const int l=lung();
  int quanti=0;
  tensor2 td(0);
  std::vector<teta> th = get_teta(*this);
  for (int i=0; i<2*l; i++) if (th[i].facce().size() == 3)
    for (int i2=i+1; i2<2*l; i2++)
      if (th[i]==th[i2] || th[i]==-th[i2])
        td.push_back(th[i].facce());
  return td;
}

//restituisce un vettore contenente gli spigoli che danno Mobius triplets
std::vector<int> ografo::spigoli_con_mobius_triplet() const {
  int cc = ff.size();
  std::vector<int> spigoli(0);
	std::vector<teta> th = get_teta(*this);
	bool mobius_triplet = FALSE;
  for (int i=0; i<th.size(); i++)
    if (th[i].pass_twice())
      spigoli.push_back(i);
  return spigoli;
}

//rimuove il vertice vert, cambiando di conseguenza quelli successivi.
void ografo::rimuovi(int vert) {
	delete v[vert]; v[vert]=0;
  v.erase(v.begin() + vert);
  const int l=lung();
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++)
      if (get_v(i,j)>vert) v[i]->set_vicino(j,get_v(i,j)-1);
}

//scrive i dati dell'ografo su f, con formato ographs se tipo_formato = 0,
//con formato SnapPea se tipo_formato = 1 (funziona solo se ha tori al bordo)
//e con formato SnapPea ristretto se tipo_formato = 2 (senza informazioni sul bordo)
void ografo::scrivi(std::ostream& f, int tipo_formato) const{
  const int l=lung();
	if (tipo_formato == 0) {
		f << "N.Vertici: ";
		f << l << "\n";
		for (int i=0; i<l ; i++) {
			f << "N. " << i;
			if (i<10) f << " ";
			f << ":   ";
			for (int j=0; j<4; j++)
				f << std::setw(2) << get_v(i,j);
			f << "    ";
			for (int j=0; j<4; j++)
				f << std::setw(2) << get_col(i,j);
			f << "    ";
			for (int j=0; j<4; j++)
				get_tor(i,j).scrivi(f);
			f << "\n";
		}
	}
	tensor2 adiac;
	if (tipo_formato == 1) {
		superficie bd = bordo(*this);
		adiac = adiacenze_bordo_su_vertice(*this);
		if (spigoli_falsi().size()) return;
		f << "% Triangulation\n";
		f << "from ographs\n";
		f << "degenerate_solution  0.00000000\n";
		f << "oriented_manifold\n";
		f << "CS_unknown\n\n";
		f << bd.componenti_connesse() << " " << 0 << "\n";
		for (int i=0; i<bd.componenti_connesse(); i++)
			f << "    torus   0.000000000000   0.000000000000\n";
		f << "\n";
	}
	if (tipo_formato == 1 || tipo_formato == 2) {
		f << lung() << "\n";
		for (int i=0; i<lung(); i++) {
			for (int j=0; j<4; j++) {
				f << "   "; if (j) f << " ";
				f << get_v(i,j);
			}
			f << "\n";
			for (int j=0; j<4; j++) {
				int perm[4];
				for (int k=0; k<4; k++) {
					if (k==j) perm[k] = get_col(i,j);
					else {
						int corsia = mod(j-k,4)-1;
						if (j % 2) corsia = 2-corsia;
						corsia = mod(corsia - get_tor(i,j).get_t(), 3);
						if (get_col(i,j) % 2) corsia = 2-corsia;
						perm[k] = mod(get_col(i,j) + corsia + 1, 4);
					}
				}
				f << " " << perm[0] << perm[1] << perm[2] << perm[3];
			}
			f << "\n";
			if (tipo_formato == 1) {
				for (int j=0; j<4; j++) {
					f << "   "; if (j) f << " ";
					f << adiac[i][(j+2)%4]-1;
				}
				f << "\n";
				for (int j=0; j<4; j++)
					f << "  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0\n";
				f << "  0   0\n\n";
			}
		}
	}
}

void ografo::scrivi_facce(std::ostream& f) const {
  f << (inglese ? "Faces: " : "Facce: ");
  for (int i=0; i<ff.size(); i++) {
    if (i) f << ", ";
    int lati = (ff[i].embedded()) ? 1:-1;
    lati*=ff[i].lati();
    f << lati;
  }
  f << "\n";
}

//permutazione delle uscite del vertice 'vert' secondo la permutazione 'z'.
//parte ereditata da igrafo + effetto sulle facce
void ografo::Cz(int vert, const int* z, int delta_tor) {
	igrafo::Cz(vert, z, delta_tor);
  int sz = ff.size();
  for (int i=0; i<sz; i++) {
		faccia& f = ff[i];
    for (int t=0; t<f.lati(); t++) if (f.get_passaggio(t)==vert) {
      f.set_uscita(t,z[f.get_uscita(t)]);
      f.set_entrata(t,z[f.get_entrata(t)]);
    }
  }
}

//permuta: parte ereditata da igrafo + effetto sulle facce
void ografo::trasforma(const trasformazione& trasf) {
  //le facce devono essere permutate prima, altrimenti le mosse C non tornano.
  int sz = ff.size();
  for (int i=0; i<sz; i++) {
		faccia& f = ff[i];
    for (int t=0; t<f.lati(); t++) 
			f.set_passaggio(t,trasf.t[0][f.get_passaggio(t)]);
	}	
	igrafo::trasforma(trasf);
}

//dice se e' chiusa o no.
bool ografo::chiusa(const faccia& f) const {
  int lati = f.lati();
  return (!falso(f.get_passaggio(lati-1), f.get_uscita(lati-1)));
}

// x==y per gli overtici
bool overtice::operator==(const vertice& y) const {
  if (!ivertice::operator==(y)) return FALSE;
  for (int t=0; t<4; t++)
    if (get_tor(t)!=y.get_tor(t)) return FALSE;
  return TRUE;
}

// x<y lessicografico su (grafo, vicini, cols);
bool ografo::operator<(const grafo& yg) const {
  const ografo& y = (ografo&) yg;
  if ((igrafo) *this < (igrafo) y) return TRUE;
  if ((igrafo) y < (igrafo) *this) return FALSE;
  const int l = lung();
	for (int i=0; i<l; i++)
    for (int j=0; j<4; j++) {
      int xtor = get_tor(i,j).get_t();
      int ytor = y.get_tor(i,j).get_t();
      if (xtor < ytor) return TRUE;
      if (xtor > ytor) return FALSE;
    }
	return FALSE;
}

//assegnamento
ografo& ografo::operator=(const ografo& o) {
	disalloca(v);
	v = std::vector<vertice*>(0);
	for (int i=0; i<o.lung(); i++)
		v.push_back(o[i]->clone());
	ff = o.get_ff();
	return *this;
}

//ritorna un vettore th, ove th[i] e' un teta contenente la triade
//di facce adiacenti allo spigolo i-esimo.
//Se 'colora_progressivo' = TRUE (default), allora prima colora le facce
//  progressivamente (partendo cioe' da 1)
//Se 'orientazioni_concordi' = TRUE (default), allora le triplette in th[i]
//  hanno tutte la stessa orientazione (indotta da orientazione del poliedro)
//  (in questo modo una tripletta e' letta [a,b,c] o [-c,-b,-a] a seconda
//  di come e' orientato lo spigolo)
//ATTENZIONE: se lo spigolo i-esimo e' falso, ed una faccia e' adiacente ad esso
//e ad un altro spigolo falso, la triple th[i] dipende dalla direzione assegnata allo spigolo 
//e non e' quindi affidabile.
std::vector<teta> get_teta(const ografo& o, bool colora_progressivo, bool orientazioni_concordi) {
  if (colora_progressivo) colora_progressivamente(o.ff);
  const int l = o.lung();
  std::vector<teta> ls(0);
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++) 
      if (o.prima_occorrenza_spigolo(i,j)) {
				teta th = teta(o.ff,i,j);
				if (orientazioni_concordi && j % 2) {
					int temp = th.get_elem(0); th.set_elem(0,th.get_elem(1)); th.set_elem(1,temp);
				}
        ls.push_back(th);
			}
  return ls;
}

//dice se i colori dati alle facce sono ammissibili
bool ografo::colori_facce_ammissibili() const {
  std::vector<teta> th = get_teta(*this, FALSE);
  const int l = lung();
  for (int i=0; i<2*l; i++)
    if (!th[i].ammissibile())
      return FALSE;
  return TRUE;
}

//costruisce le facce di 'o'
std::vector<faccia> facce(const ografo& o) {
  std::vector<faccia> ff(0);
  const int l=o.lung(); 
	tensor3 w(l,tensor2(4,std::vector<int>(3,0)));
	tensor2 th2(l,std::vector<int>(4,0));
	int passaggio, entrata, uscita;
  for (int i=0; i<l; i++) {
    for (int j=0; j<4; j++) {
      for (int c=0; c<3; c++) {
        if (!w[i][j][2-c]) {
					faccia nuova;  //cambiato 14/09
          ff.push_back(nuova);
          faccia& f = ff.back();
          f.passaggio.push_back(i);
          f.entrata.push_back(j);
          f.uscita.push_back(svincolo[j][c]);
          int corsia=c;
          w[i][f.entrata[0]][2-corsia]=1;
          w[i][f.uscita[0]][corsia]=1; 
          int fine=0; bool invertita=FALSE;
          for (int k=0; !fine ; k++){
            int p = f.get_passaggio(k);
            torsione tors = o.get_tor(p,f.get_uscita(k));
            while (!fine && tors.falsa()) {
              if (invertita==TRUE) fine = TRUE;
              else {
                invertita = TRUE; f.inverti();
                p = f.get_passaggio(k);
                corsia=2-c;
                tors = o.get_tor(p,f.get_uscita(k));
              }
            }
            if (!fine) {
              passaggio = o.get_v(p,f.uscita[k]);

              entrata = o.get_col(p,f.uscita[k]);
              
              if (mod(f.get_uscita(k),2)) corsia=2-corsia;
              corsia=mod(corsia - tors.get_t(),3);
              if (mod(entrata,2)) corsia=2-corsia;
              
              uscita=svincolo[entrata][corsia];

              w[passaggio][entrata][2-corsia]=1;
              w[passaggio][uscita][corsia]=1;
              if (passaggio==f.get_passaggio(0) && uscita==f.get_uscita(0) && corsia==c) 
								fine=1;
              else f.add_passaggio(passaggio, uscita, entrata);
            }
          }
        }
      }
    }
  }
	return ff;
}


